#pragma once

#include <iostream>
#include <map>
#include <algorithm>

// 此处重载了<<操作符，在ColMeta中进行了调用
template<typename T, typename = typename std::enable_if<std::is_enum<T>::value, T>::type>
std::ostream &operator<<(std::ostream &os, const T &enum_val) {
    os << static_cast<int>(enum_val);
    return os;
}

template<typename T, typename = typename std::enable_if<std::is_enum<T>::value, T>::type>
std::istream &operator>>(std::istream &is, T &enum_val) {
    int int_val;
    is >> int_val;
    enum_val = static_cast<T>(int_val);
    return is;
}

struct Rid {
    int page_no;
    int slot_no;

    friend bool operator==(const Rid &x, const Rid &y) {
        return x.page_no == y.page_no && x.slot_no == y.slot_no;
    }

    friend bool operator!=(const Rid &x, const Rid &y) {
        return !(x == y);
    }

    // 重载小于运算符以用于比较
    friend bool operator<(const Rid &x, const Rid &y) {
        if (x.page_no != y.page_no) {
            return x.page_no < y.page_no;
        }
        return x.slot_no < y.slot_no;
    }

    // 重载小于运算符以用于比较
    friend bool operator>(const Rid &x, const Rid &y) {
        if (x.page_no != y.page_no) {
            return x.page_no > y.page_no;
        }
        return x.slot_no > y.slot_no;
    }

    // 重载小于等于运算符以用于比较
    friend bool operator<=(const Rid &x, const Rid &y) {
        if (x.page_no != y.page_no) {
            return x.page_no < y.page_no;
        }
        return x.slot_no <= y.slot_no;
    }

    // 重载大于等于运算符以用于比较
    friend bool operator>=(const Rid &x, const Rid &y) {
        if (x.page_no != y.page_no) {
            return x.page_no > y.page_no;
        }
        return x.slot_no >= y.slot_no;
    }

    // 静态方法，返回无效的Rid
    static Rid Invalid() {
        return Rid{-1, -1};
    }
};

/* 定义字段类型枚举 */
enum ColType {
    TYPE_INT, TYPE_FLOAT, TYPE_STRING
};

/* 字段类型转字符串函数 */
inline std::string coltype2str(ColType type) {
    std::map<ColType, std::string> m = {
            {TYPE_INT,    "INT"},
            {TYPE_FLOAT,  "FLOAT"},
            {TYPE_STRING, "STRING"}
    };
    return m.at(type);
}

// 判断两列的数据类型是否能够比较
inline bool checkType(ColType lhs, ColType rhs) {
    if ((lhs == TYPE_INT && rhs == TYPE_FLOAT) || (lhs == TYPE_FLOAT && rhs == TYPE_INT) ||
        (lhs == TYPE_STRING && rhs == TYPE_STRING)) {
        return true;
    } else {
        return lhs == rhs;
    }
}

class RecScan {
public:
    virtual ~RecScan() = default;

    virtual void next() = 0;

    virtual bool is_end() const = 0;

    virtual Rid rid() const = 0;
};